<?php
/**
 * @author Jan Haller <jan.haller@math.uzh.ch>
 */

use IMATHUZH\Qfq\Core\Report\Link;
use IMATHUZH\Qfq\Core\Store\Store;
use IMATHUZH\Qfq\Core\Database\Database;

/**
 * Creates a DOMNodeList from the HTML content and iterates over each element
 * If the content is empty, a link gets created to add content to the wiki
 * The first element gets preceded by a link to edit the content as a whole
 * <h1> to <h6> tags get preceded by a link to edit the section including its child elements
 * Macros ({{toc}}, {{childPages}}, [[wikiPage#Heading]], [[pageSlug/wikiPage#Heading]], {{collapse()}}, {{collapse}}) will be replaced
 * Output to the report is generated by the echo commands
 *
 * @param $paramArray
 * @param $qfq
 * @return void
 * @throws CodeException
 * @throws DbException
 * @throws UserFormException
 * @throws UserReportException
 */
function renderWiki($paramArray, $qfq) {
    $headingTags = array(HTML_TAG_H1, HTML_TAG_H2, HTML_TAG_H3, HTML_TAG_H4, HTML_TAG_H5, HTML_TAG_H6);
    $unwantedTags = array(HTML_TAG_HTML, HTML_TAG_BODY);

    // id of the wiki page record
    $id = $paramArray[WIKI_PAGE_COLUMN_ID];

    // Will be either '|r:0' or '|r:5'
    $renderMode = '|r:' . $paramArray[WIKI_RENDER_MODE];

    // Will be either true or false
    $printMode = (isset($paramArray[WIKI_PRINT_MODE])) ? $paramArray[WIKI_PRINT_MODE] : false;

    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_CONTENT . ', ' . WIKI_PAGE_COLUMN_PAGE_SLUG . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_ID . ' = ' . $id ;
    $resultArray = doSql($sql);

    $wikiPageContent = (isset($resultArray[0][WIKI_PAGE_COLUMN_CONTENT])) ? $resultArray[0][WIKI_PAGE_COLUMN_CONTENT] : '';
    $pageSlug = (isset($resultArray[0][WIKI_PAGE_COLUMN_PAGE_SLUG])) ? $resultArray[0][WIKI_PAGE_COLUMN_PAGE_SLUG] : '';

    // This is necessary in print mode to hide the button text
    $buttonText = ($printMode) ? '' : '|t:Edit';

    // No wiki page has been created yet
    if (empty($wikiPageContent) && empty($pageSlug)) {
        return;

    // Wiki page has been created but is without content
    // Creates an 'edit' link and exits the function
    } elseif (empty($wikiPageContent)) {
        echo '<div style="float:right;">';
        echo doQfqLink('p:', $pageSlug, '&form=wikiEditor&r=0&wpIdPrevious=' . $id . '&action=new' . $renderMode . '|s|b|E' . $buttonText);
        echo '&nbsp;</div>';
        echo '<div class="wiki-navigation" style="display:inline-block; margin-bottom:10px">' . doNavigation($id) . '</div>';
        return;
    }

    // Navigation
    echo ($printMode) ? '' : '<div class="wiki-navigation">' . doNavigation($id) . '</div>';

    // Will be passed as a parameter in the link and enables the edit of the content as a whole
    // Always used for the first element
    $action = WIKI_TOKEN_EDIT_ALL;

    // This DOM will be updated with additional elements
    $dom = new DomDocument();
    $dom->loadHTML($wikiPageContent);

    // Used for iterating over the elements
    $xpath = new DOMXPath($dom);
    $content = $xpath->query('//*');

    $counter = 1;
    $firstElement = true;

    // Iterate over each element
    foreach ($content as $element) {
        $nodeName = $element->nodeName;

        // The first two elements are always <html> and <body>
        if (!in_array($nodeName, $unwantedTags)) {
            $nodeValue = $element->nodeValue;

            // Looks for <h1> - <h6> elements or the first element
            // Creates a <div> containing the edit link and an <a>-tag
            // In printMode this will be skipped
            if ((in_array($nodeName, $headingTags) || $firstElement) && !$printMode) {

                // If the first element is not <h1> - <h6>, then $sectionId = 0
                $sectionId = (in_array($nodeName, $headingTags)) ? $counter : 0;

                // <div>
                $divHtml = '<div title="Edit this section" id="wiki-section-' . $sectionId . '" style="float:right;">';
                $divHtml .= doQfqLink('p:', $pageSlug, '&form=wikiEditor&r=0&sectionId=' . $sectionId . '&wpIdPrevious=' . $id . '&action=' . $action . $renderMode . '|s|b|E' . $buttonText);
                $divHtml .= (in_array($nodeName, $headingTags)) ? '<a name="' . slugify($nodeValue) . '"></a>&nbsp;</div>' : '&nbsp;</div>';

                // Creates a DOMElement from HTML and inserts it before the current element into the DOM
                $divElement = doDomElement($divHtml);
                $divElement = $dom->importNode($divElement, true);
                $parentNode = $element->parentNode;
                $parentNode = $parentNode->insertBefore($divElement, $element);

                $counter = ($sectionId === 0) ? $counter : $counter + 1;
                unset($divHtml, $divElement, $parentNode, $sectionId);
            }

            // Reset values
            $buttonText = '';
            $action = WIKI_TOKEN_EDIT_SECTION;
            $firstElement = false;

            // Check for macro '{{toc}}'
            if (trim($nodeValue) === WIKI_TOKEN_TABLE_OF_CONTENTS) {

                // HTML for the list table of contents
                $tocHtml = doTableOfContents($content);

                // Check if list table of contents has been successfully created
                if ($tocHtml !== $nodeValue) {

                    // Creates a DOMElement from HTML and inserts it before the current element into the DOM
                    $ulElement = doDomElement($tocHtml);
                    $ulElement = $dom->importNode($ulElement, true);
                    $parentNode = $element->parentNode;
                    $parentNode = $parentNode->replaceChild($ulElement, $element);

                    unset($tocHtml, $ulElement, $parentNode);
                }
            }

            // Check for macro '{{childPages}}'
            if (trim($nodeValue) === WIKI_TOKEN_CHILD_PAGES) {

                // HTML for the list child pages
                $cpHtml = doChildPages($id);

                // Check if list child pages has been successfully created
                if ($cpHtml !== $nodeValue) {

                    // Creates a DOMElement from HTML and inserts it before the current element into the DOM
                    $ulElement = doDomElement($cpHtml);
                    $ulElement = $dom->importNode($ulElement, true);
                    $parentNode = $element->parentNode;
                    $parentNode = $parentNode->replaceChild($ulElement, $element);

                    unset($cpHtml, $ulElement, $parentNode);
                }
            }

            // Check for and replace macro '[[wikiPage#Heading]]' and '[[pageSlug/wikiPage#Heading]]'
            if (preg_match('<(\[\[[a-zA-Z0-9- :,;/\-]+#{1}[a-zA-Z0-9. :,;\-]*\]\])>', $nodeValue)) {
                $linkHtml = preg_replace_callback('<(\[\[[a-zA-Z0-9- :,;/\-]+#{1}[a-zA-z0-9. :,;\-]*\]\])>', 'doWikiLink', $nodeValue);

                // Check if link has been successfully created
                if ($linkHtml !== $nodeValue) {

                    // Creates a DOMElement from HTML and inserts it before the current element into the DOM
                    $aElement = doDomElement($linkHtml);
                    $aElement = $dom->importNode($aElement, true);
                    $parentNode = $element->parentNode;
                    $parentNode = $parentNode->replaceChild($aElement, $element);

                    unset($linkHtml, $aElement, $parentNode);
                }
            }
        }
    }

    // Raw HTML
    $wikiHtml = $dom->saveHTML();

    // HTML is different in print mode
    if ($printMode) {

        // Remove {{collapse()}} and {{collapse}}
        // This keeps the content visible on the PDF
        $wikiHtml = preg_replace('<<[a-zA-Z]+>{{collapse\((.*?)\)}}<[a-zA-Z/]+>>', '', $wikiHtml);
        $wikiHtml = preg_replace('<(<[a-zA-Z]+>{{collapse}}<[a-zA-Z/]+>)>', '', $wikiHtml);
    } else {

        // Search and replace opening collapse macro
        // Opening <div>-tag gets inserted which acts as a parent for the collapsed content
        // Search: <p>{{collapse(optional text)}}</p>
        // Replace: <a><span></span>optional text</a></br><div>
        $wikiHtml = preg_replace_callback('<<[a-zA-Z]+>{{collapse\((.*?)\)}}<[a-zA-Z/]+>>', 'doCollapse', $wikiHtml);

        // Search and replace closing collapse macro
        // Close previously opened <div>-tag
        // Search: <p>{{collapse}}</p>
        // Replace: </div>
        $wikiHtml = preg_replace('<(<[a-zA-Z]+>{{collapse}}<[a-zA-Z/]+>)>', '</div>', $wikiHtml);
    }

    // Output
    echo $wikiHtml;
}

/**
 * Creates a DOMNodeList from the HTML content and iterates over each element
 * Output will only contain the sections that will be loaded into the editor
 *
 * @param $paramArray
 * @param $qfq
 * @return void
 */
function loadWikiContent($paramArray, $qfq) {
    $headingTags = array(HTML_TAG_H1, HTML_TAG_H2, HTML_TAG_H3, HTML_TAG_H4, HTML_TAG_H5, HTML_TAG_H6);

    // id of the wiki page record
    $id = $paramArray[WIKI_PAGE_COLUMN_ID];

    // id of the section that will be edited
    $sectionId = intval($paramArray[WIKI_PAGE_SECTION_ID]);

    // Will be either 'editAll' or 'editSection'
    $action = $paramArray[WIKI_TOKEN_ACTION];

    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_CONTENT . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_ID . ' = ' . $id;
    $resultArray = doSql($sql);

    $wikiPageContent = $resultArray[0][WIKI_PAGE_COLUMN_CONTENT];

    // If the content as a whole will be edited there is no need to parse
    if ($action === WIKI_TOKEN_EDIT_ALL) {
        echo $wikiPageContent;
        return;
    }

    $nodeLevelSection = null;
    $nodeLevelCurrent = null;

    $counter = 1;
    $dom = new DOMDocument();
    $dom->loadHTML($wikiPageContent);
    $body = $dom->getElementsByTagName(HTML_TAG_BODY)->item(0);
    $content = $body->childNodes;

    // Iterate over each element
    foreach ($content as $element) {
        $nodeName = $element->nodeName;

        // Looks for <h1> - <h6> elements
        if (in_array($nodeName, $headingTags)) {

            // Look for the element to edit
            if ($counter === $sectionId) {

                // Extract the level from the element
                // The level consist of the number from the heading tag, e.g. 1 for <h1>
                $nodeLevelSection = intval(substr($nodeName, 1));
                $nodeLevelCurrent = intval(substr($nodeName, 1)) + 1;

                // Slugified heading gets set in S-Store to later use it as an anchor in page forward
                setVar(WIKI_PAGE_HEADING_TITLE, slugify($element->nodeValue), STORE_SIP);

            // Element is not editable
            } elseif ($counter < $sectionId) {
                $counter++;
                continue;

            // Element is editable
            } else {

                // Extract the level from the element
                $nodeLevelCurrent = intval(substr($nodeName, 1));
            }

            // Element is not editable
            if ($nodeLevelSection >= $nodeLevelCurrent) {
                break;
            }
            $counter++;
        }

        // Check if element is editable
        if ($nodeLevelSection < $nodeLevelCurrent) {

            // Output element
            echo $dom->saveHTML($element);
        }
    }

    // Current section gets set in S-Store to later extract the unedited section of the content
    setVar(WIKI_PAGE_SECTION_ID_END, $counter, STORE_SIP);
}

/**
 * Creates a DOMNodeList from the HTML content and iterates over each element
 * Return will only contain the sections that were not edited
 *
 * @param $paramArray
 * @param $qfq
 * @return null[]|string[]
 */
function getWikiContentSection($paramArray, $qfq) {
    $headingTags = array(HTML_TAG_H1, HTML_TAG_H2, HTML_TAG_H3, HTML_TAG_H4, HTML_TAG_H5, HTML_TAG_H6);

    // id of the wiki page record
    $id = $paramArray[WIKI_PAGE_COLUMN_ID];

    // Will be a positive number
    $sectionIdStart = intval($paramArray[WIKI_PAGE_SECTION_ID_START]);

    // Will either be true or false
    // Necessary to output non-heading elements
    $includeTag = ($paramArray[WIKI_TOKEN_INCLUDE_TAG] === "true") ? true : false;
    $section = null;

    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_CONTENT . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_ID . ' = ' . $id;
    $resultArray = doSql($sql);

    $wikiPageContent = $resultArray[0][WIKI_PAGE_COLUMN_CONTENT];

    $dom = new DOMDocument();
    $dom->loadHTML($wikiPageContent);
    $body = $dom->getElementsByTagName(HTML_TAG_BODY)->item(0);
    $content = $body->childNodes;

    // If not set, that means the section is until the end of the content
    $sectionIdEnd = (isset($paramArray[WIKI_PAGE_SECTION_ID_END])) ? intval($paramArray[WIKI_PAGE_SECTION_ID_END]) : count($content) + 1;
    $counter = 1;

    // Iterate over each element
    foreach ($content as $element) {
        $nodeName = $element->nodeName;

            // Looks for <h1> - <h6> elements
            if (in_array($nodeName, $headingTags)) {

                // Checks if the following elements should be included, e.g. <p>, <table>, etc.
                if ($counter >= $sectionIdStart && $counter < $sectionIdEnd) {

                    // HTML of the element is added to $section
                    $section .= $dom->saveHTML($element);
                    $includeTag = true;
                } else {
                    $includeTag = false;
                }

                $counter++;

            } elseif($includeTag) {

                // HTML of the element is added to $section
                $section .= $dom->saveHTML($element);
            }
        }

    // Current section gets set in V-Store, so it can later be used to update the current wiki page content
    return [WIKI_TOKEN_SECTION => $section];
}

/**
 * Creates a DOMElement from given HTML
 *
 * @param $html
 * @return mixed DOMElement
 */
function doDomElement($html) {
    $dom = new DOMDocument();
    $element = $dom->loadHTML($html);
    $element = $dom->getElementsByTagName(HTML_TAG_BODY)->item(0)->firstChild;

    return $element;
}

/**
 * Recursive function to generate a navigation with links from the current page to its parent page and so on.
 * Finished version looks like this: grandparent page > parent page > current page
 *
 * @param $id
 * @return string
 * @throws CodeException
 * @throws DbException
 * @throws UserFormException
 * @throws UserReportException
 */
function doNavigation($id) {
    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_WP_ID_PARENT . ', ' . WIKI_PAGE_COLUMN_PAGE_SLUG . ', ' . WIKI_PAGE_COLUMN_NAME . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_ID . ' = ' . $id ;
    $resultArray = doSql($sql);

    // If no $wpIdParent is found this equals to null
    $wpIdParent = $resultArray[0][WIKI_PAGE_COLUMN_WP_ID_PARENT];
    $pageSlug = $resultArray[0][WIKI_PAGE_COLUMN_PAGE_SLUG];
    $name = $resultArray[0][WIKI_PAGE_COLUMN_NAME];

    $link = doQfqLink('p:', $pageSlug, '&wpId=' . $id . '|t:' . $name);
    $separator = '';
    $output = '';

    // Check if the root page has been reached
    if (!is_null($wpIdParent)) {
        $output .= doNavigation($wpIdParent);
        $separator = '<span>&nbsp;»&nbsp;</span>';
    }

    $output .= $separator . $link;
    return $output;
}

/**
 * Creates a table of contents from all headings of the wiki page
 * Gets invoked by the macro '{{toc}}'
 * Every level/element has its own indentation
 *
 * @param $content
 * @return string
 */
function doTableOfContents($content) {
    $headingTags = array(HTML_TAG_H1, HTML_TAG_H2, HTML_TAG_H3, HTML_TAG_H4, HTML_TAG_H5, HTML_TAG_H6);
    $headings = array();

    // <h6>
    $highestLevel = 6;

    // <h1>, lowest level
    $previousLevel = 1;
    $toc = WIKI_TOKEN_TABLE_OF_CONTENTS;

    // Iterate over each element
    foreach($content as $element) {
        $nodeName = $element->nodeName;

        // Looks for <h1> - <h6> elements
        if (in_array($nodeName, $headingTags)) {
            $nodeValue = $element->nodeValue;

            // Heading level and name get added to the array
            $headings[] = intval(substr($nodeName, 1)) . "=" . $nodeValue;
        }
    }

    // If no heading elements were found the macro '{{toc}}' will be returned as is
    if (empty($headings)) {
        return $toc;
    }

    // List for the table of contents
    $toc = '<ul class="wiki-toc"><li><strong>Table of contents</strong></li>';

    // Iterate over each heading
    foreach($headings as $key => $value) {
        $headingLevel = explode("=", $value);
        $currentLevel = $headingLevel[0];
        $heading = $headingLevel[1];

        // Check if current level is higher than the highest level, update if true
        $highestLevel = ($currentLevel < $highestLevel) ? $currentLevel : $highestLevel;

        if ($currentLevel > $previousLevel) {

            // 'Indentation' of the <ul> element inside the table of contents
            $toc .= str_repeat('<ul>', $currentLevel - $previousLevel);

        } elseif ($currentLevel === $previousLevel) {
            $toc .= '</li>';
        } else {

            // Closing the 'indented' <ul> element inside the table of contents
            $toc .= str_repeat('</ul>', $previousLevel - $currentLevel);
        }

        $toc .= '<li>';

        // Anchor tag that links to the heading
        $toc .= '<a href="#' . slugify($heading) . '">' . $heading . '</a>';
        $previousLevel = $currentLevel;
    }

    // Complete the list
    $toc .= str_repeat('</li></ul>', $previousLevel);
    return $toc;
}

/**
 * Creates a link to a wiki page within the same wiki or in another wiki
 * Gets invoked by the macros '[[wikiPage#Heading]]' and/or '[[pageSlug/wikiPage#Heading]]'
 *
 * @param $match
 * @return string
 */
function doWikiLink($match) {
    // e.g. $match = '[[wikiPage#Heading]]'
    // e.g. $match = '[[pageSlug/wikiPage#Heading]]'

    // Removes '[[' and ']]'
    $match = substr($match[0], 2, -2);

    // Separates 'wikiPage'/'pageSlug/wikiPage' and 'Heading'
    $arr = explode("#", $match);
    $wikiPage = $arr[0];
    $heading = $arr[1];

    // Looks for the name of the wiki page: '[[wikiPage#Heading]]'
    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_ID . ', ' . WIKI_PAGE_COLUMN_PAGE_SLUG . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_NAME . ' = "' . $wikiPage . '" AND ISNULL(' .WIKI_PAGE_COLUMN_WP_ID_CURRENT . ')';
    $resultArray = doSql($sql);

    // Looks for the combination of page slug and name of the wiki page: '[[pageSlug/wikiPage#Heading]]'
    if (empty($resultArray)) {
        $sql = 'SELECT ' . WIKI_PAGE_COLUMN_ID . ', ' . WIKI_PAGE_COLUMN_PAGE_SLUG .  ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE CONCAT(' . WIKI_PAGE_COLUMN_PAGE_SLUG . ', "/",' . WIKI_PAGE_COLUMN_NAME . ') = "' . $wikiPage . '" AND ISNULL(' . WIKI_PAGE_COLUMN_WP_ID_CURRENT . ')';
        $resultArray = doSql($sql);
    }

    // If no wiki page was found the macro will be returned as is
    if (empty($resultArray)) {
        return '[[' . $match . ']]';
    }

    // id of the wiki page
    $id = $resultArray[0][WIKI_PAGE_COLUMN_ID];

    // page slug of the wiki page
    $pageSlug = $resultArray[0][WIKI_PAGE_COLUMN_PAGE_SLUG];

    // Link to the wiki page with anchor to the heading
    $link = doQfqLink('p:', $pageSlug, '&wpId=' . $id . '#' . slugify($heading) . '|t:' . $heading . '|c:wiki-link');
    return $link;
}

/**
 * Creates a list of child pages of the current wiki page
 * Gets invoked by the macro '{{childPages}}'
 *
 * @param $id
 * @return string
 */
function doChildPages($id) {
    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_ID . ', ' . WIKI_PAGE_COLUMN_NAME . ', ' . WIKI_PAGE_COLUMN_PAGE_SLUG . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_WP_ID_PARENT . ' = ' . $id . ' AND ISNULL(' . WIKI_PAGE_COLUMN_WP_ID_CURRENT . ') ORDER BY ' . WIKI_PAGE_COLUMN_NAME . ' ASC';
    $resultArray = doSql($sql);

    $childPages = '{{childPages}}';

    // If no child pages were found the macro gets returnd as is
    if (empty($resultArray)) {
        return $childPages;
    }

    // List of the child pages
    $childPages = '<ul class="wiki-pages-hierarchy">';

    // Iterate over each row
    foreach ($resultArray as $row) {
        $id = $row[WIKI_PAGE_COLUMN_ID];
        $name = $row[WIKI_PAGE_COLUMN_NAME];
        $pageSlug = $row[WIKI_PAGE_COLUMN_PAGE_SLUG];

        // Link to the child page
        $childPages .= '<li>' . doQfqLink('p:', $pageSlug, '&wpId=' . $id . '|t:' . $name) . '</li>';
    }

    // Complete the list
    $childPages .= '</ul>';
    return $childPages;
}

/**
 * Checks the privileges of the user that wants to access the wiki page
 *
 * @param $paramArray
 * @param $qfq
 * @return string[]
 * @throws CodeException
 * @throws DbException
 * @throws UserFormException
 * @throws UserReportException
 */
function checkWikiAccess($paramArray, $qfq) {
    $store = Store::getInstance();

    // id of the wiki page
    $id = $paramArray[WIKI_PAGE_COLUMN_ID];

    $sql = 'SELECT ' . WIKI_PAGE_COLUMN_AUTHOR . ', ' . WIKI_PAGE_COLUMN_RO_USER . ', ' . WIKI_PAGE_COLUMN_RO_GROUP . ', ' . WIKI_PAGE_COLUMN_RO_PUBLIC . ', ' . WIKI_PAGE_COLUMN_RW_USER . ', ' . WIKI_PAGE_COLUMN_RW_GROUP . ', ' . WIKI_PAGE_COLUMN_RW_PUBLIC . ', ' . WIKI_PAGE_COLUMN_PAGE_LOCK . ' FROM ' . WIKI_PAGE_TABLE_NAME . ' WHERE ' . WIKI_PAGE_COLUMN_ID . ' = ' . $id ;
    $resultArray = doSql($sql);

    $author = (isset($resultArray[0][WIKI_PAGE_COLUMN_AUTHOR])) ? $resultArray[0][WIKI_PAGE_COLUMN_AUTHOR] : '';
    $roUser = (isset($resultArray[0][WIKI_PAGE_COLUMN_RO_USER])) ? explode(',', $resultArray[0][WIKI_PAGE_COLUMN_RO_USER]) : array();
    $roGroup = (isset($resultArray[0][WIKI_PAGE_COLUMN_RO_GROUP])) ? explode(',', $resultArray[0][WIKI_PAGE_COLUMN_RO_GROUP]) : array();
    $roPublic = (isset($resultArray[0][WIKI_PAGE_COLUMN_RO_PUBLIC])) ? $resultArray[0][WIKI_PAGE_COLUMN_RO_PUBLIC] : '';
    $rwUser = (isset($resultArray[0][WIKI_PAGE_COLUMN_RW_USER])) ? explode(',', $resultArray[0][WIKI_PAGE_COLUMN_RW_USER]) : array();
    $rwGroup = (isset($resultArray[0][WIKI_PAGE_COLUMN_RW_GROUP])) ? explode(',', $resultArray[0][WIKI_PAGE_COLUMN_RW_GROUP]) : array();
    $rwPublic = (isset($resultArray[0][WIKI_PAGE_COLUMN_RW_PUBLIC])) ? $resultArray[0][WIKI_PAGE_COLUMN_RW_PUBLIC] : '';
    $pageLock = (isset($resultArray[0][WIKI_PAGE_COLUMN_PAGE_LOCK])) ? $resultArray[0][WIKI_PAGE_COLUMN_PAGE_LOCK] : '';

    // Get logged-in user from T-store
    $feUser = $store::getVar(TYPO3_FE_USER, STORE_TYPO3);

    // Get groups of logged-in user from T-store
    $feUserGroup[] = explode(',', $store::getVar(TYPO3_FE_USER_GROUP, STORE_TYPO3));

    $feUserGroup = (is_array($feUserGroup)) ? $feUserGroup : array();
    $wikiAccess = WIKI_TOKEN_ACCESS_OFF;

    /* Checks if permissions are set to read-write for either the public, the logged-in user or a group of the logged-in user
       Also check if the logged-in user is the author of the wiki page */
    if ($rwPublic === WIKI_TOKEN_ACCESS_ON || (in_array($feUser, $rwUser) && !empty($feUser))  || (array_intersect($feUserGroup[0], $rwGroup) && !in_array('', $feUserGroup[0])) || $feUser === $author) {
        $wikiAccess = WIKI_TOKEN_ACCESS_READ_WRITE;

        // Checks if permissions are set to read-only for either the public, the logged-in user or a group of the logged-in user
    } elseif ($roPublic === WIKI_TOKEN_ACCESS_ON || (in_array($feUser, $roUser) && !empty($feUser))  || (array_intersect($feUserGroup[0], $roGroup) && !in_array('', $feUserGroup[0]))) {
        $wikiAccess = WIKI_TOKEN_ACCESS_READ_ONLY;
    }

    // If page lock is set but not by the logged-in user who has read-write permissions, they change to read-only
    $wikiAccess = (!empty($pageLock) && $wikiAccess === WIKI_TOKEN_ACCESS_READ_WRITE && $pageLock !== $feUser) ? WIKI_TOKEN_ACCESS_READ_ONLY : $wikiAccess;

    // Permissions get set in V-Store, so they can later be used to render the wiki accordingly
    return [WIKI_TOKEN_WIKI_ACCESS => $wikiAccess];
}

/**
 * Replaces the macro '{{collapse()}}' and/or '{{collapse(string)}}'
 *
 * @param $matches
 * @return mixed|string
 */
function doCollapse($matches) {
    // $matches[0] = '<p>{{collapse(optional text)}}</p>
    // $matches[1] = 'optional text'
    $text = $matches[1];
    $randomInt = rand();
    $str = '<a class="wiki-collapse-link" style="cursor:pointer;" onclick="$(\'#wiki-collapse-chevron-'. $randomInt . '\').toggleClass(\'glyphicon-chevron-down glyphicon-chevron-right\'); $(\'#wiki-collapse-content-' . $randomInt . '\').fadeToggle(\'fast\');">
            <span class="wiki-collapse-chevron glyphicon glyphicon-chevron-right" id="wiki-collapse-chevron-' . $randomInt . '"></span>' . $text . '</a></br>
            <div class="wiki-collapse-container" style="display: none;" id="wiki-collapse-content-' . $randomInt . '">';

    return $str;
}

/**
 * Executes SQL query
 *
 * @param $sql
 * @return array|bool|int|string
 * @throws CodeException
 * @throws DbException
 * @throws UserFormException
 */
function doSql($sql) {
    $db = new Database();
    $result = $db->sql($sql);
    return $result;
}

/**
 * Creates a QFQ-link
 *
 * @param $page
 * @param $pageSlug
 * @param $params
 * @return string
 * @throws CodeException
 * @throws DbException
 * @throws UserFormException
 * @throws UserReportException
 */
function doQfqLink($page, $pageSlug, $params) {
    $s = Store::getInstance();
    $sip = $s->getSipInstance();
    $link = new Link($sip);
    $str = $link->renderLink($page . $pageSlug . $params);
    return $str;
}

/**
 * Sets a variable to a store
 *
 * @param $key
 * @param $value
 * @param $store
 * @return void
 * @throws CodeException
 * @throws UserFormException
 * @throws UserReportException
 */
function setVar($key, $value, $store) {
    $s = Store::getInstance();
    $s::setVar($key, $value, $store);
}

/**
 * Slugify function
 * E.g. This is my custom heading -> This-is-my-custom-heading
 *
 * @param $text
 * @return array|string|string[]
 */
function slugify($text) {
    $divider = '-';

    // Replace non letter or digits by divider
    $text = preg_replace('~[^\pL\d]+~u', $divider, $text);

    // Transliterate
    $text = iconv('utf-8', 'us-ascii//TRANSLIT', $text);

    // Remove unwanted characters
    $text = preg_replace('~[^-\w]+~', '', $text);

    // Trim
    $text = trim($text, $divider);

    // Remove duplicate divider
    $text = preg_replace('~-+~', $divider, $text);

    return $text;
}